use std::collections::BTreeMap;

use itertools::Itertools;

use crate::data::{Root, SourceFile, Span, Spanned};

use super::{LedgerEntry, Transaction};

fn print_transaction(
    target: &mut dyn std::io::Write,
    root: &Root,
    tx: &Transaction,
    padding: usize,
) -> std::io::Result<()> {
    writeln!(
        target,
        "{}: {}",
        tx.datestamp,
        tx.title.as_deref().unwrap_or("")
    )?;

    if !tx.annotations.is_empty() {
        writeln!(
            target,
            "    {}",
            tx.annotations.iter().map(|a| format!("[{a}]")).join(" ")
        )?;
    }

    let force_show = tx.changes.iter().unique_by(|b| *b.account).count() > 1;

    for change in &tx.changes {
        let spacer = if change.amount.is_sign_positive() {
            " "
        } else {
            ""
        };

        let balance = if let Some(bal) = change.balance {
            format!(" = {bal}")
        } else {
            String::new()
        };

        if Some(*change.unit) == root.account_spec(*change.account).and_then(|s| s.unit)
            && !force_show
        {
            writeln!(
                target,
                " - {:padding$}: {spacer}{}{balance}",
                change.account, change.amount,
            )?;
        } else {
            writeln!(
                target,
                " - {:padding$}: {spacer}{}{balance} {}",
                change.account, change.amount, change.unit,
            )?;
        }
    }
    // empty line afterwards
    writeln!(target)
}

fn print_comment(target: &mut dyn std::io::Write, c: &Spanned<String>) -> std::io::Result<()> {
    writeln!(target, "{c}")
}

pub fn print_ledger<'l>(
    fsdata: &mut crate::io::FilesystemData,
    root: &Root,
    entries: impl Iterator<Item = &'l LedgerEntry>,
) -> std::io::Result<()> {
    let mut ordering = BTreeMap::<Option<SourceFile>, BTreeMap<Span, Vec<&LedgerEntry>>>::new();

    entries.for_each(|e| {
        ordering
            .entry(e.span().context)
            .or_default()
            .entry(e.span())
            .or_default()
            .push(e);
    });

    let Some(padding) = root.spec_root.accounts.keys().map(|k| k.len()).max() else {
        // no accounts
        return Ok(());
    };

    for (source, entries) in ordering {
        let Some(filename) = source else {
            log::warn!("Skipping ledger content with no source file!");
            continue;
        };
        log::info!(
            "Updating ledger file {} ...",
            std::path::Path::new(filename.as_str()).display()
        );

        let mut outfile = fsdata.try_write(filename)?;

        for (_span, mut les) in entries {
            les.sort_by_key(|le| le.as_transaction().map(|tx| tx.datestamp));
            for le in les {
                match le {
                    LedgerEntry::Transaction(tx) => {
                        print_transaction(&mut outfile, root, tx, padding)?
                    }
                    LedgerEntry::Comment(c) => print_comment(&mut outfile, c)?,
                }
            }
        }

        outfile.commit()?;
    }
    Ok(())
}
