use std::collections::HashMap;

pub struct SourceTag;
impl stringstore::NamespaceTag for SourceTag {
    const PREFIX: &'static str = "loc";
}
/// Canonicalized path to a file
pub type SourceFile = stringstore::StoredOsString<SourceTag>;

/// Helper for accessing data on the filesystem
#[derive(Default)]
pub struct FilesystemData {
    file_data: HashMap<SourceFile, ariadne::Source<std::rc::Rc<str>>>,
}

impl FilesystemData {
    pub fn try_write(&self, file: SourceFile) -> std::io::Result<FileOutput> {
        let data_path = std::path::Path::new(file.as_str()).parent().unwrap();
        let tmpfile = tempfile::Builder::new()
            .prefix(".")
            .suffix(".tmp")
            .tempfile_in(data_path)?;

        Ok(FileOutput {
            target: file,
            tmpfile,
        })
    }
}

pub struct FileOutput {
    target: SourceFile,
    tmpfile: tempfile::NamedTempFile<std::fs::File>,
}

impl FileOutput {
    pub fn cancel(self) -> std::io::Result<()> {
        self.tmpfile.close()
    }
    pub fn commit(self) -> std::io::Result<()> {
        self.tmpfile.persist(self.target.as_str())?;
        Ok(())
    }
}

impl std::io::Write for FileOutput {
    fn write(&mut self, buf: &[u8]) -> std::io::Result<usize> {
        self.tmpfile.write(buf)
    }
    fn flush(&mut self) -> std::io::Result<()> {
        self.tmpfile.flush()
    }
}

impl std::fmt::Debug for FilesystemData {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("FilesystemData").finish()
    }
}

impl ariadne::Cache<SourceFile> for FilesystemData {
    type Storage = std::rc::Rc<str>;
    fn fetch(
        &mut self,
        id: &SourceFile,
    ) -> Result<&ariadne::Source<Self::Storage>, impl std::fmt::Debug> {
        if !self.file_data.contains_key(id) {
            match std::fs::read_to_string(id.as_str()) {
                Ok(data) => {
                    let data: std::rc::Rc<_> = std::rc::Rc::from(data.into_boxed_str());
                    self.file_data.insert(*id, ariadne::Source::from(data));
                }
                Err(e) => return Err(e),
            }
        }

        Ok(self.file_data.get(id).unwrap())
    }

    fn display<'a>(&self, id: &'a SourceFile) -> Option<impl std::fmt::Display + 'a> {
        Some(id.as_str().to_string_lossy())
    }
}
